/*
 * Copyright 2002-2003, Tobias Huerlimann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netinet/ip_var.h>
#include <netinet/tcp_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/icmp_var.h>
#include <sys/sysctl.h>

#define INCL_BASE
#define INCL_DOSERRORS
#define INCL_DOS
#define INCL_DOSMISC
#define INCL_DOSDEVICES
#define INCL_DOSDEVIOCTL
#define INCL_DOSPROFILE
#define INCL_KBD
#define INCL_LONGLONG
#include <os2.h>

#include "config.h"
#include "bsdlibc.h"
#include "rndos2.h"
#include "add_randomness.h"

/* DosPerfSysCall() found its way into the toolkit header files
 * but not the required CMDs and structs */
#define CMD_KI_RDCNT	(0x63)
typedef struct _CPUUTIL {
	ULONG ulTimeLow;	/* Low 32 bits of time stamp      */
	ULONG ulTimeHigh;	/* High 32 bits of time stamp     */
	ULONG ulIdleLow;	/* Low 32 bits of idle time       */
	ULONG ulIdleHigh;	/* High 32 bits of idle time      */
	ULONG ulBusyLow;	/* Low 32 bits of busy time       */
	ULONG ulBusyHigh;	/* High 32 bits of busy time      */
	ULONG ulIntrLow;	/* Low 32 bits of interrupt time  */
	ULONG ulIntrHigh;	/* High 32 bits of interrupt time */
} CPUUTIL;
typedef CPUUTIL		*PCPUUTIL;



typedef struct _struct_tcpip_stats {
	struct tcpstat		my_tcpstat;
	struct udpstat		my_udpstat;
	struct icmpstat		my_icmpstat;
} struct_tcpip_stats;

typedef struct _struct_input_stats {
	struct MOU_GETPTRPOS_data_packet {
		USHORT y;
		USHORT x;
	}			mou_pos;
	KBDKEYINFO		kbd_info;
} struct_input_stats;

typedef struct _struct_system_stats {
	ULONG			sys_info[QSV_FOREGROUND_PROCESS - QSV_MS_COUNT + 1];
} struct_system_stats;

TID		add_randomness_tid = NULLHANDLE;

int add_to_pool(PVOID data, ULONG data_len);
int query_tcpip(void);
int query_input(void);
int query_system(void);

int add_to_pool(PVOID data, ULONG data_len) {
	int rc;

	rc = rndos2_write(RND_SRND, data, data_len);
	if (rc != 0)
		return -1;
	rc = rndos2_ioctl(RND_SRND, RNDADDTOENTCNT, &data_len);
	return (rc == 0 ? 0 : -1);
}

/* Queries TCP/IP statistics */
int query_tcpip(void) {
	int			rc;
	int			err = 0;
	int			mib[7] = {0, 0, 0, 0, 0, 0, 0};
	struct_tcpip_stats	stats;
	size_t			buf_len;
	unsigned long int	size_in_ints = (sizeof(stats) / sizeof(int));
	unsigned long int	iloop;

	mib[0] = CTL_NET;
	mib[1] = PF_INET;

	/* Get TCP statistics */
	mib[2] = IPPROTO_TCP;
	mib[3] = TCPCTL_STATS;
	buf_len = sizeof(stats.my_tcpstat);
	rc = sysctl((int*)&mib, 4, &stats.my_tcpstat, &buf_len, NULL, 0);
	if (rc == -1) {
		psock_errno("rndos2: sysctl()");
		err = -1;
	}

	/* Get UDP statistics */
	mib[2] = IPPROTO_UDP;
	mib[3] = UDPCTL_STATS;
	buf_len = sizeof(stats.my_udpstat);
	rc = sysctl((int*)&mib, 4, &stats.my_udpstat, &buf_len, NULL, 0);
	if (rc == -1) {
		psock_errno("rndos2: sysctl()");
		err = -1;
	}

#if 0
	/* Get ICMP statistics */
	mib[2] = IPPROTO_ICMP;
	mib[3] = ICMPCTL_STATS;
	buf_len = sizeof(stats.my_icmpstat);
	rc = sysctl((int*)&mib, 4, &stats.my_icmpstat, &buf_len, NULL, 0);
	if (rc == -1) {
		psock_errno("rndos2: sysctl()");
		err = -1;
	}
#endif

//	for (iloop = 0; iloop < size_in_ints; iloop++) {
//		add_net_randomness(*(int*)((ULONG)&stats + iloop*4));
//	}
	
	return add_to_pool(&stats, sizeof(stats));
}



/* Queries user input like pointer position or keyboard state */
int query_input(void) {
	APIRET			rc = NO_ERROR;
	int			rc2;
	int			err = 0;
	struct_input_stats	stats;
	HFILE			dev_handle;
	ULONG			open_action;
	SHORT			kbd_parm;
	ULONG			parm_len;
	ULONG			data_len;
	unsigned long int	size_in_ints = (sizeof(stats) / sizeof(int));
	unsigned long int	iloop;

	/* Get pointer position */
	rc = DosOpen("MOUSE$", &dev_handle, &open_action, 0UL, FILE_NORMAL,
		FILE_OPEN, OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYNONE, NULL);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosOpen() failed (rc = %u)\n", rc);
		err = -1;
	}
	parm_len = 0;
	data_len = sizeof(stats.mou_pos);
	rc = DosDevIOCtl(dev_handle, IOCTL_POINTINGDEVICE, MOU_GETPTRPOS, NULL, parm_len,
		&parm_len, &stats.mou_pos, data_len, &data_len);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosDevIOCtl() failed (rc = %u)\n", rc);
		return 1;
	}
	rc = DosClose(dev_handle);

#if 0
	/* Get keyboard status */
	rc = DosOpen("KBD$", &dev_handle, &open_action, 0UL, FILE_NORMAL,
		FILE_OPEN, OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYNONE, NULL);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosOpen() failed (rc = %u)\n", rc);
		err = -1;
	}
	parm_len = sizeof(kbd_parm);
	data_len = sizeof(stats.kbd_info);
	rc = DosDevIOCtl(dev_handle, IOCTL_KEYBOARD, KBD_PEEKCHAR, &kbd_parm,
		parm_len, &parm_len, &stats.kbd_info, data_len, &data_len);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosDevIOCtl() failed (rc = %u)\n", rc);
		return 1;
	}
	rc = DosClose(dev_handle);
#endif

	return add_to_pool(&stats, sizeof(stats));
}



/* Queries system information and statistics */
int query_system() {
	APIRET			rc = NO_ERROR;
	int			rc2;
	int			err = 0;
	struct_system_stats	stats;
	unsigned long int	size_in_ints = (sizeof(stats) / sizeof(int));
	unsigned long int	iloop;
	ULONG			num_cpus;
	QSPTRREC*		qss_data;
	ULONG			qss_start, qss_stop, qss_size;
	CPUUTIL			*cpu_util;

	/* SysInfo: not extremly random ... */
	rc = DosQuerySysInfo(QSV_MS_COUNT, QSV_FOREGROUND_PROCESS,
		(PVOID)&stats.sys_info, sizeof(stats.sys_info));
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosQuerySysInfo() failed (rc = %u)\n", rc);
		return rc;
	}

	/* Query the number of processors, allocate that many CPUUTIL structs and
	 * perform a DosPerfSysCall() for CPU stats */
	rc = DosQuerySysInfo(QSV_NUMPROCESSORS, QSV_NUMPROCESSORS, (PVOID)&num_cpus, sizeof(num_cpus));
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosQuerySysInfo() failed (rc = %u)\n", rc);
		return rc;
	}
	cpu_util = (CPUUTIL*)malloc(sizeof(CPUUTIL) * num_cpus);
	if (cpu_util == NULL) {
		fprintf(stderr, "rndos2: malloc() failed\n");
		return 1;
	}
	rc = DosPerfSysCall(CMD_KI_RDCNT, (ULONG)cpu_util, 0UL, 0UL);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosPerfSysCall() failed (rc = %u)\n", rc);
		return rc;
	}
	add_to_pool(cpu_util, sizeof(CPUUTIL) * num_cpus);
	free((PVOID)cpu_util);

	/*
	 * Use the mostly undocumented API DosQuerySysState() to get
	 * information about processes, threads, modules, semaphores,
	 * open files, ...
	 */
	/* First, we need a quite large data buffer */
	rc = DosAllocMem((PPVOID)&qss_data, QSS_BUFSIZE, PAG_COMMIT | PAG_READ | PAG_WRITE);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosAllocMem() failed (rc = %u)\n", rc);
		return rc;
	}
	bzero(qss_data, QSS_BUFSIZE);
	/* Using QS_MODVER in combination with a too small buffer size might
	 * result in an IPE with not-so-super-recent kernels */
	rc = DosQuerySysState(QS_PROCESS | QS_SEMAPHORE | QS_MTE | QS_FILESYS | QS_SHMEMORY,
		0UL, (PID)0, (TID)0, qss_data, QSS_BUFSIZE);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosQuerySysState() failed (rc = %u)\n", rc);
		return rc;
	}
	/*
	 * Now we need to guess the size of the useable part of the
	 * buffer. The following size calculation is a quick'n'dirty
	 * hack:
	 * a)   It requires the system wide shared memory information
	 *      to be the last the data buffer returned by
	 *      DosQuerySysState()
	 * b)	It assumes that there are no "gaps" in the data buffer
	 * c)	System wide shared memory information is omitted
	 * These assumptions are based on tests but not on any
	 * information in the Toolkit Reference. There is no guarantee
	 * that this won't change in future API implementations.
	 */
	qss_start = (ULONG)qss_data->pGlobalRec;
	qss_stop = (ULONG)qss_data->pMemRec - 1;
	qss_size = qss_stop - qss_start + 1;
	add_to_pool(qss_data, qss_size);
	rc = DosFreeMem(qss_data);
	if (rc != NO_ERROR) {
		fprintf(stderr, "rndos2: DosFreeMem() failed (rc = %u)\n", rc);
	}

	return add_to_pool(&stats, sizeof(stats));
}



void APIENTRY add_randomness_thread(ULONG param_ptr) {
	APIRET			rc = NO_ERROR;
	struct add_randomness_conf	conf = *((struct add_randomness_conf*)param_ptr);

	while (TRUE) {
		if (conf.flags & AR_QUERY_SYSTEM)
			query_system();
		if (conf.flags & AR_QUERY_CON)
			query_input();
		if (conf.flags & AR_QUERY_TCPIP)
			query_tcpip();
		rc = DosSleep(conf.interval);
	}
}
