/*
 * Copyright 2002, Tobias Huerlimann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define INCL_ERRORS
#define INCL_DOSPROFILE
#include <os2.h>

#define BUFSIZE		(128*1024)

#define HF_STDIN	((HFILE)0)
#define HF_STDOUT	((HFILE)1)
#define HF_STDERR	((HFILE)2)

char		buf[BUFSIZE];
QSPTRREC*	top = (QSPTRREC*)&buf;

void dump_GlobalRec(void);
void dump_ProcRec(void);
void dump_MemRec(void);
void dump_LibRec(void);
void dump_FSRec(void);

main(void) {
	APIRET		rc = NO_ERROR;
	PVOID		end;

	memset(&buf, 0, sizeof(buf));

	rc = DosQuerySysState(QS_SUPPORTED,
		0UL, (PID)0, (TID)0, &buf, sizeof(buf));

	if (rc != NO_ERROR) {
		fprintf(stderr, "DosQuerySysState() failed (rc = %u)\n", rc);
		return rc;
	}

	printf("Start:\t%u\n", (ULONG)top->pGlobalRec);
	printf("Stop:\t%u\n", (ULONG)top->pMemRec);
	
	printf("\t-> Size: \t%u\n", (ULONG)top->pMemRec - (ULONG)top->pGlobalRec);

	printf("pGlobalRec:\t%u\n",	(ULONG)top->pGlobalRec);
	printf("pProcRec:\t%u\n",	(ULONG)top->pProcRec);
	printf("p16SemRec:\t%u\n",	(ULONG)top->p16SemRec);
	printf("p32SemRec:\t%u\n",	(ULONG)top->p32SemRec);
	printf("pMemRec:\t%u\n",	(ULONG)top->pMemRec);
	printf("pLibRec:\t%u\n",	(ULONG)top->pLibRec);
	printf("pShrMemRec:\t%u\n",	(ULONG)top->pShrMemRec);
	printf("pFSRec:\t%u\n",		(ULONG)top->pFSRec);

	dump_GlobalRec();
//	dump_ProcRec();
//	dump_MemRec();
	dump_LibRec();
	dump_FSRec();

	return 0;
}

void dump_GlobalRec() {
	APIRET		rc;
	ULONG		start, stop, len;
	ULONG		written;

	printf("## GlobalRec\n");
	start = (ULONG)top->pGlobalRec;
	stop = start + sizeof(QSGREC) - 1;
	len = stop - start + 1;
	printf("\tStart:\t%u\n\tStop:\t%u\n\tLen:\t%u\n", start, stop, len);
#ifndef NODUMP
	printf("\tDump:\t");fflush(stdout);
	rc = DosWrite(HF_STDOUT, (PVOID)start, len, &written);
#endif
	printf("\n\n");fflush(stdout);
}


void dump_ProcRec() {
	APIRET		rc;
	ULONG		start, stop, len;
	QSPREC*		my;
	ULONG		written;

	printf("## ProcRec\n");
	start = (ULONG)top->pProcRec;
	my = (QSPREC*)top->pProcRec;
//	while (my->RecType == 1) {
//		stop = (ULONG)my + sizeof(QSPREC) + sizeof(QS
//	}
	stop = start + sizeof(QSGREC) - 1;
	len = stop - start + 1;
	printf("\tStart:\t%u\n\tStop:\t%u\n\tLen:\t%u\n", start, stop, len);
#ifndef NODUMP
	printf("\tDump:\t");fflush(stdout);
	rc = DosWrite(HF_STDOUT, (PVOID)start, len, &written);
#endif
	printf("\n\n");fflush(stdout);
}


void dump_MemRec() {
	APIRET		rc;
	ULONG		start, stop, len;
	QSMREC*		my;
	ULONG		written;

	printf("## MemRec\n");
	start = (ULONG)top->pMemRec;
	my = (QSMREC*)top->pMemRec;
	while (my->MemNextRec != 0) {
		my += my->MemNextRec;
	}
	stop = (ULONG)my + sizeof(QSMREC) - 1;
	len = stop - start + 1;
	printf("\tStart:\t%u\n\tStop:\t%u\n\tLen:\t%u\n", start, stop, len);
#ifndef NODUMP
	printf("\tDump:\t");fflush(stdout);
	rc = DosWrite(HF_STDOUT, (PVOID)start, len, &written);
#endif
	printf("\n\n");fflush(stdout);
}


void dump_LibRec() {
	APIRET		rc;
	ULONG		start, stop, len;
	QSLREC*		my;
	ULONG		written;

	printf("## LibRec\n");
	start = (ULONG)top->pLibRec;
	my = (QSLREC*)top->pLibRec;
	while (my->pNextRec) {
		my = my->pNextRec;
	}
	/* Ignore module pathname and possibly following object records
	 * to keep it simple */
	stop = (ULONG)my->pName;
	len = stop - start + 1;
	printf("\tStart:\t%u\n\tStop:\t%u\n\tLen:\t%u\n", start, stop, len);
#ifndef NODUMP
	printf("\tDump:\t");fflush(stdout);
	rc = DosWrite(HF_STDOUT, (PVOID)start, len, &written);
#endif
	printf("\n\n");fflush(stdout);
}


void dump_FSRec() {
	APIRET		rc;
	ULONG		start, stop, len;
	QSFREC*		my;
	ULONG		written;

	printf("## FSRec\n");
	start = (ULONG)top->pFSRec;
	my = (QSFREC*)top->pFSRec;
	while (my->RecType != 0) {
		my = my->pNextRec;
	}
	stop = (ULONG)my - 1;
	len = stop - start + 1;
	printf("\tStart:\t%u\n\tStop:\t%u\n\tLen:\t%u\n", start, stop, len);
#ifndef NODUMP
	printf("\tDump:\t");fflush(stdout);
	rc = DosWrite(HF_STDOUT, (PVOID)start, len, &written);
#endif
	printf("\n\n");fflush(stdout);
}

