/*
 * Copyright 2002, Tobias Huerlimann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Very incomplete replacement for OpenBSD's timeout_* kernel API */

#ifdef RNDEBUG
	#include <stdio.h>	/* For fprintf() */
#endif

#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#include <os2.h>

#include "timeout.h"

void APIENTRY timeout_thread(ULONG param) {
	APIRET		rc = NO_ERROR;
	struct timeout*	to = (struct timeout*)param;
	void		(*my_func)(void*) = to->to_func;
	void*		my_arg = to->to_arg;
	ULONG		my_time = to->to_time;

	DosSleep(my_time);
	to->to_flags |= TIMEOUT_TRIGGERED;
	my_func(my_arg);
	to->to_flags &= ~TIMEOUT_ONQUEUE;
	to->to_thread = NULLHANDLE;
}

void timeout_set(struct timeout* to, void (*fn)(void*), void* arg) {
	to->to_func = fn;
	to->to_arg = arg;
	to->to_time = 0;
	to->to_flags = TIMEOUT_INITIALIZED;
	to->to_thread = NULLHANDLE;
}

void timeout_add(struct timeout* to, int timeout_time) {
	APIRET		rc = NO_ERROR;

	if (!(to->to_flags & TIMEOUT_INITIALIZED)) {
#ifdef RNDEBUG
		fprintf(stderr, "rndos2: timeout_add(): timeout not initialized\n");
#endif
		return;
	}
	if (to->to_flags & TIMEOUT_ONQUEUE) {
#ifdef RNDEBUG
		fprintf(stderr, "rndos2: timeout_add(): timeout already queued\n");
#endif
		return;
	}

	to->to_time = timeout_time;
	to->to_flags |= TIMEOUT_ONQUEUE;
	to->to_flags &= ~TIMEOUT_TRIGGERED;

	rc = DosCreateThread(&to->to_thread, &timeout_thread,
		(ULONG)to, CREATE_READY | STACK_SPARSE, 16384UL);
	if (rc != NO_ERROR) {
#ifdef RNDEBUG
		fprintf(stderr, "rndos2: DosCreateThread() failed (rc = %u)\n", rc);
#endif
	}
}

void timeout_del(struct timeout* to) {
	APIRET		rc = NO_ERROR;

	/* Return if no timeout has been scheduled */
	if (!(to->to_flags & TIMEOUT_ONQUEUE) | (to->to_thread == NULLHANDLE))
		return;

	/* Return if the timeout has already been reached and the
	 * function is already running or has been returned
	 */
	if (to->to_flags & TIMEOUT_TRIGGERED)
		return;

	rc = DosKillThread(to->to_thread);
	if (rc != NO_ERROR) {
#ifdef RNDEBUG
		fprintf(stderr, "rndos2: DosKillThread() failed (rc = %u)\n", rc);
#endif
	}

	to->to_flags &= ~TIMEOUT_ONQUEUE;
	to->to_thread = NULLHANDLE;
}

